/*
    Copyright (C) 2003 Evan Buswell
    
    Based almost completely on:
    mpg321 - a fully free clone of mpg123.
    ao.c: Copyright (C) 2001, 2002 Joe Drew
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "fakeao.h"
#include <sys/ioctl.h>
#include <mad.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/soundcard.h>

char *devicename = "/dev/dsp";

ao_device *playdevice = NULL;

void open_ao_playdevice(struct mad_header const *header) {
    int tempfd;
    int format = AFMT_S16_NE;
    int channels;
    unsigned int rate;
    if(playdevice == NULL) {
	playdevice = malloc(sizeof(ao_device));
	if(playdevice == NULL) {
	    perror("Error in open_ao_playdevice");
	    exit(1);
	}
	playdevice->file = NULL;
    }
    if(playdevice->file != NULL) {
	tempfd = fileno(playdevice->file);
	if(ioctl(tempfd, SNDCTL_DSP_RESET) == -1) {
	    perror("Error in open_ao_playdevice");
	    exit(1);
	}
    } else {
	playdevice->file = fopen(devicename, "a");
	if(playdevice->file == NULL) {
	    perror("Error in open_ao_playdevice");
	    exit(1);
	}
	tempfd = fileno(playdevice->file);
    }
    /* sample format */
    if(ioctl(tempfd, SNDCTL_DSP_SETFMT, &format) == -1) {
	perror("Error in open_ao_playdevice");
	exit(1);
    }
    /* number of channels */
    channels = header->mode > 0 ? 2 : 1;
    if(ioctl(tempfd, SNDCTL_DSP_CHANNELS, &channels) == -1) {
	perror("Error in open_ao_playdevice");
	exit(1);
    }
    /* sampling rate */
    rate = header->samplerate;
    if(ioctl(tempfd, SNDCTL_DSP_SPEED, &rate) == -1) {
	perror("Error in open_ao_playdevice");
	exit(1);
    }
}

void ao_play(ao_device *thedevice, unsigned char *stream, int length) {
    if((fwrite(stream, 1, length, thedevice->file) != length)
       && (ferror(thedevice->file))) {
	fprintf(stderr, "Failed to write to output device.\n");
	exit(1);
    }
    if(fflush(thedevice->file) == -1) {
	perror("Error in ao_play");
	exit(1);
    }
}
