/*
    Copyright (C) 2003 Evan Buswell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>

#include "../common/port.h"
#include "timer.h"
#include "data.h"

struct sockaddr_in baddr;
int b_sock;

void b_senddata(int current) {
    static unsigned char *buffer = NULL;
    if(buffer == NULL)
	buffer = (unsigned char *) malloc(MAXFRAMESIZE);

    if(frames[current].length > MAXFRAMESIZE) {
	fprintf(stderr, "Error: strange frame size\n");
	return;
    }
    
    if(frames[current].buffer + frames[current].length > bigbuf.fake_end) {
	memcpy(buffer, frames[current].buffer,
	       (bigbuf.fake_end - frames[current].buffer) + 1);
	memcpy(buffer + (bigbuf.fake_end - frames[current].buffer) + 1,
	       bigbuf.buffer,
	       frames[current].length -
	       (bigbuf.fake_end - frames[current].buffer) + 1);
	sendto(b_sock, buffer, frames[current].length, 0,
	       (struct sockaddr *)&baddr, sizeof(struct sockaddr_in));
    } else {
	sendto(b_sock, frames[current].buffer, frames[current].length, 0,
	       (struct sockaddr *)&baddr, sizeof(struct sockaddr_in));
    }
}

int b_init(void) {
    int on = 1;
    int priority = 65535;
    struct sockaddr_in addr;
    
    /* bind a socket... */
    b_sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(b_sock == -1)
	return(-1);
    
    if(setsockopt(b_sock, SOL_SOCKET, SO_BROADCAST, &on, sizeof(int)) == -1)
	return(-1);
    if(setsockopt(b_sock, SOL_SOCKET, SO_PRIORITY, &priority, sizeof(int)) == -1) {
	/* try with a lower value */
	priority = 6;
	setsockopt(b_sock, SOL_SOCKET, SO_PRIORITY, &priority, sizeof(int)); /* (ignore errors) */
    }

    
    addr.sin_family = AF_INET;
    addr.sin_port = htons(BROADCASTPORT_D);
    addr.sin_addr.s_addr = (u_int32_t) INADDR_ANY;
    
    baddr.sin_family = AF_INET;
    baddr.sin_port = htons(BROADCASTPORT_C);
    baddr.sin_addr.s_addr = (u_int32_t) INADDR_BROADCAST;
    
    if(bind(b_sock, (struct sockaddr *) &addr, sizeof(struct sockaddr_in)) == -1)
	return(-1);

    return(0);
}
