/*
    Copyright (C) 2003 Evan Buswell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <pthread.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <mad.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>

#include "timer.h"
#include "data.h"
#include "../common/frame.h"
#include "local.h"
#include "../common/port.h"

pthread_t l_thread;

int l_getlistener(int lsock) {
    int s;
    fprintf(stderr, "Listening for local provider on port %d...\n", LISTENPORT);
    listen(lsock, 1); /* We only get one connection on this port.
			 Everything else is refused. */
    while((s = accept(lsock, NULL, NULL)) == -1) {
	sleep(1);
	/* retry after error.  Really bad errors shouldn't happen. */
    }
    fprintf(stderr, "Provider connected.\n");
    return(s);
}


void *l_listen(void *arg) {
    int i;
    int s;
    int lsock;
    unsigned char *current;
    int wrap = 0;
    int junk = 0;
    struct mad_header header;
    struct mad_stream stream;
    uid_t uid;
    char *oob;
    struct timespec waittime;
    unsigned char junkbuf[512];
    waittime.tv_sec = 0;
    waittime.tv_nsec = 250000000; /* a quarter second */

    oob = (char *) malloc(32);

    lsock = *((int *) arg);
    free(arg);
    
    /* drop priveleges */
    uid = getuid();
    seteuid(uid);
    
    current = bigbuf.buffer;

 listen_start_loop:
    s = l_getlistener(lsock);
    while(1) {
	if((recv(s, oob, 32, MSG_OOB | MSG_DONTWAIT | MSG_NOSIGNAL) > 0)
	   && strncmp(oob, "flush", 5)) {
	    /* test if we would have needed to wrap */
	    if((current + bigbuf.mtu + MAD_BUFFER_GUARD)
	       > (bigbuf.buffer + bigbuf.length)) {
		bigbuf.fake_end = current - 1;
		current = bigbuf.buffer;
	    }
	    bigbuf.top = current;
	    bigbuf.remaining = bigbuf.length - (bigbuf.top - bigbuf.buffer);
	}

	i = recv(s, bigbuf.top, bigbuf.remaining, MSG_NOSIGNAL);
	if(i == -1 || i == 0) {
	    if((i == -1) && (errno != EAGAIN) && (errno != EINTR)) {
		if(junk != 0) {
		    fprintf(stderr, "%d bytes\n", junk);
		    junk = 0;
		}
		/* wait for a new connection */
		perror("Lost provider connection");
		close(s);
		current = bigbuf.top;
		goto listen_start_loop;
	    }
	    /* see if provider is still there */
	    i = sock_write_bytes(s, "ping", 4); /* we get tcp acks, so there's no need to send
						   a pong from the provider */
	    if((i == -1) && (errno != EINTR)) {
		if(junk != 0) {
		    fprintf(stderr, "%d bytes\n", junk);
		    junk = 0;
		}
		/* wait for a new connection */
		perror("Lost provider connection");
		close(s);
		goto listen_start_loop;
	    }
	    nanosleep(&waittime, NULL);
	} else {
	    bigbuf.top += i;
	    bigbuf.remaining -= i;
	    if((bigbuf.top + bigbuf.mtu + MAD_BUFFER_GUARD) > (bigbuf.buffer + bigbuf.length)) {
		/* copy MAD_BUFFER_GUARD bytes to the beginning and decrement fake_end */
		bigbuf.fake_end = bigbuf.top - 1; /* this is the last written byte */
		bigbuf.fake_end -= (MAD_BUFFER_GUARD - 1);
		memcpy(bigbuf.buffer, bigbuf.fake_end + 1, (MAD_BUFFER_GUARD -1));
		bigbuf.top = bigbuf.buffer + (MAD_BUFFER_GUARD - 1);
		bigbuf.remaining = bigbuf.length - (MAD_BUFFER_GUARD - 1);
		wrap = 1;
	    }
	}
	
	/* search for frame separator */
	while(!wrap
	      ? (current + MAD_BUFFER_GUARD < bigbuf.top)
	      : wrap) {
	    if(testboundary(current, 0 /* offset */)) {
		if(junk != 0) {
		    fprintf(stderr, "%d bytes\n", junk);
		    junk = 0;
		}
		mad_header_init(&header);
		mad_stream_init(&stream);
		i = decode_next_offset(current, MAD_BUFFER_GUARD, /* why make this real? */
				       &header, &stream, 0 /* offset */);
		if(i == -1) {
		    current++;
		    fprintf(stderr, "Problem decoding provider stream: %s\nContinuing...\n",
			    mad_stream_errorstr(&stream));
		} else {
		    t_addtostream(current, i, header); /* we pass header by VALUE! */
		    mad_stream_finish(&stream);
		    mad_header_finish(&header); /* this actually does nothing anyway,
						   so we don't worry about prematurely
						   freeing crap. */
		    current += i;
		}
	    } else {
		if(junk == 0) {
		    fprintf(stderr, "Junk in provider stream...");
		    fflush(stderr);
		}
		junk++;
		current++;
	    }
	    if(wrap && (current > bigbuf.fake_end)) {
		current = (bigbuf.buffer + (current - bigbuf.fake_end)) - 1;
		wrap = 0;
	    }
	    if((recv(s, oob, 32, MSG_OOB | MSG_DONTWAIT | MSG_NOSIGNAL) > 0)
	       && strncmp(oob, "flush", 5)) {
		/* test if we would have needed to wrap */
		if((current + bigbuf.mtu + MAD_BUFFER_GUARD)
		   > (bigbuf.buffer + bigbuf.length)) {
		    bigbuf.fake_end = current - 1;
		    current = bigbuf.buffer;
		}
		bigbuf.top = current;
		bigbuf.remaining = bigbuf.length - (bigbuf.top - bigbuf.buffer);
	    }
	}
    }
}

int l_init(void) {
    int *s;
    int r;
    int on = 1;
    size_t mtu = DEFAULT_MTU;
    
    struct sockaddr_in addr;
    
    /* bind a socket to listen on */
    s = (int *) malloc(sizeof(int));
    if(s == NULL)
	return(-1);
    *s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(*s == -1)
	return(-1);

    if(setsockopt(*s, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(int)) == -1)
	return(-1);
    if(setsockopt(*s, SOL_SOCKET, SO_RCVBUF, &mtu, sizeof(size_t)) == -1)
	return(-1);
    
    addr.sin_family = AF_INET;
    addr.sin_port = htons(LISTENPORT);
    addr.sin_addr.s_addr = (u_int32_t) INADDR_ANY;
    
    if(bind(*s, (struct sockaddr *) &addr, sizeof(struct sockaddr_in)) == -1)
	return(-1);
    
    /* start the thread */
    r = pthread_create(&l_thread, NULL, l_listen, s);
    return(r);
}
