/*
    Copyright (C) 2003 Evan Buswell
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <pthread.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <mad.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>

#include "bcast.h"
#include "timer.h"
#include "data.h"
#include "bcast.h"

pthread_t t_thread;

int t_current = 0;

/*#define sendtobcast(buffer, length)  write(mpipe[1], buffer, length)*/

volatile int receivedalarm = 0;
volatile int receivedwakeup = 0;
volatile int timerasleep = 0;

pthread_mutex_t tsleepmutex = PTHREAD_MUTEX_INITIALIZER;

void sigalarmfunc(int whocares) {
    if(signal(SIGALRM, SIG_IGN) == SIG_ERR) {
	perror("signal(SIGALRM, SIG_IGN");
    }
    receivedalarm = 1;
}

void sigusr1func(int whocares) {
    if(signal(SIGUSR1, SIG_IGN) == SIG_ERR) {
	perror("signal(SIGUSR1, SIG_IGN)");
    }
    receivedwakeup = 1;
}

/* this is actually executed in the local listener thread */
int t_addtostream(unsigned char *buffer, int length, struct mad_header header) {
    static int current;
    static int sentwake = 0; /* this is to try and avoid sending a signal in the middle of the
				handler routine */
    struct timespec waittime = {
	tv_sec: 0,
	tv_nsec: 250000000
    }; /* a quarter second */

    while(current < t_current ?
	  t_current - current < 10 :
	  (t_current + NUMFRAMES) - current < 10) {
	/* We're approaching wraparound; cool off a bit */
	nanosleep(&waittime, NULL);
    }

    frames[current].duration.tv_sec = header.duration.seconds;
    frames[current].duration.tv_usec = mad_timer_fraction(header.duration, 1000000);
                                                                        /* ^ in microseconds */

    if(frames[current].duration.tv_sec == 0 &&
       frames[current].duration.tv_usec == 0) {
	/* ignore this frame; it's probably bad */
	return(0);
    }
    
    frames[(current + 1) % NUMFRAMES].buffer = NULL;

    frames[current].length = length;
    frames[current].buffer = buffer;

    current = (current + 1) % NUMFRAMES;
    
    pthread_mutex_lock(&tsleepmutex);
    if(timerasleep) {
	if(!sentwake) {
	    fprintf(stderr, "Waking timer\n");
	    pthread_kill(t_thread, SIGUSR1);
	    sentwake = 1;
	}
    } else {
	sentwake = 0;
    }
    pthread_mutex_unlock(&tsleepmutex);
    return(0);
}


void *t_stream(void *arg) { /* arg is unused */
    struct itimerval mytimer;
    uid_t uid;
    sigset_t newmask;
#ifdef DEBUG
    struct timeval now, diff;
    struct timeval prev;
    prev.tv_sec = 0;
    prev.tv_usec = 0;
#endif /* DEBUG */

    signal(SIGALRM, sigalarmfunc);
    signal(SIGUSR1, sigusr1func);
    sigemptyset(&newmask);
    sigaddset(&newmask, SIGALRM);
    sigaddset(&newmask, SIGUSR1);
    pthread_sigmask(SIG_UNBLOCK, &newmask, NULL);
    
    /* drop priveleges */
    uid = getuid();
    seteuid(uid);
    
    mytimer.it_value.tv_sec = 0;
    mytimer.it_value.tv_usec = 0;
    mytimer.it_interval.tv_sec = 0;
    mytimer.it_interval.tv_usec = 0;
    
    while(1) {
	pthread_mutex_lock(&tsleepmutex);
	while(frames[t_current].buffer != NULL) {
	    pthread_mutex_unlock(&tsleepmutex);
#ifdef DEBUG
	    gettimeofday(&now, NULL);
	    if(prev.tv_sec != 0) {
		timersub(&now, &prev, &diff);
		fprintf(stderr, "waited %d secs %d usecs, meant ", diff.tv_sec, diff.tv_usec);
		fprintf(stderr, "%d secs, %d usecs.\n", mytimer.it_value.tv_sec, mytimer.it_value.tv_usec);
	    }
	    prev = now;
#endif /* DEBUG */
	    if(mytimer.it_value.tv_sec != frames[t_current].duration.tv_sec ||
	       mytimer.it_value.tv_usec != frames[t_current].duration.tv_usec) {
		mytimer.it_value = frames[t_current].duration;
		mytimer.it_interval = frames[t_current].duration;
		setitimer(ITIMER_REAL, &mytimer, NULL);
	    }

	    b_senddata(t_current);

	    while(!receivedalarm) {
		pause(); /* wait for when it's time to send the next frame; */
	    }
	    receivedalarm = 0;
	    /* re-install signal handler */
	    if(signal(SIGALRM, sigalarmfunc) == SIG_ERR) {
		perror("signal(SIGALRM, sigalarmfunc)");
	    }
	    t_current = (t_current + 1) % NUMFRAMES;
	    receivedwakeup = 0;
	    pthread_mutex_lock(&tsleepmutex);
	}
	timerasleep = 1;
	pthread_mutex_unlock(&tsleepmutex);
	fprintf(stderr, "timer:Waiting for data...(asleep)\n");
	if(signal(SIGUSR1, sigusr1func) == SIG_ERR) {
	    perror("signal(SIGUSR1, sigusr1func)");
	}
	while(!receivedwakeup) {
	    pause(); /* wait for somebody to put something in the framequeue */
	}
	/* re-install signal handler */
	signal(SIGUSR1, sigusr1func);
	fprintf(stderr, "timer:Woke up!\n");
	timerasleep = 0;
    }
}

int t_init(void) {
    int i;
    int r;
    int *j;
    struct sched_param *p;
    p = (struct sched_param *) malloc(sizeof(struct sched_param));
    j = (int *) malloc(sizeof(int));

    /* start the timer thread */
    r = pthread_create(&t_thread, NULL, t_stream, NULL);
    if(r != 0)
	return(r);
    p->sched_priority = sched_get_priority_max(SCHED_FIFO);
    p->sched_priority--;
    pthread_setschedparam(t_thread, SCHED_FIFO, p);
    pthread_getschedparam(t_thread, j, p);
    fprintf(stderr, "Timer priority set to %d, ", p->sched_priority);
    if(*j != SCHED_FIFO)
	fprintf(stderr, "SCHED_FIFO not set.\n");
    else
	fprintf(stderr, "SCHED_FIFO set.\n");
    
    return(r);
}
