/*
    Copyright (C) 2003 Evan Buswell
    
    Uses code from:
    mpg321 - a fully free clone of mpg123.
    ao.c: Copyright (C) 2001, 2002 Joe Drew
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <signal.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "../common/port.h"
#include "../common/nethelp.h"

int nextsong = 0;

void sigintfunc(int whocares) {
    signal(SIGINT, SIG_DFL);
    nextsong = 1;
    alarm(1);
}

void sigalarmfunc(int whocares) {
    signal(SIGINT, sigintfunc);
    signal(SIGALRM, sigalarmfunc);
}

void usage(char *progname) {
    fprintf(stderr, "Usage: %s [-z] [-h] file1.mp3 [file2.mp3 ...]\n\n", progname);
    fprintf(stderr, "  -z		randomize files\n");
    fprintf(stderr, "  -h		help (this message)\n");
}

/* this is adapted from mpg321 (c) 2001 Joe Drew */
void shuffle_files(char *files[], int numfiles) {
    int i;
    int a, b;
    char *swap;

    srandom(time(NULL));

    for (i = 0; i < 100 * numfiles; i++) {
        a = numfiles * ((double)random()/RAND_MAX);
        b = numfiles * ((double)random()/RAND_MAX);

        /* because if random()/RAND_MAX == 1, we will be 
           referencing an index outside the array */
        if (a == numfiles)
            a--;
        if (b == numfiles)
            b--;

        swap = files[a];
        files[a] = files[b];
        files[b] = swap;
    }
}

void flush_remote(int s) {
    send(s, "flush", 5, MSG_OOB);
}

void stream_file(int s, char *filename) {
    int filed;
    
    static char buff[4096];
    int l;
    
    filed = open(filename, O_RDONLY);
    if(filed == -1) {
	fprintf(stderr, "Failed to open %s:", filename);
	perror("");
    }
    printf("Throwing %s to local listener\n", filename);
    while(!nextsong && (l = read(filed, buff, 4096)) > 0) {
	if(sock_write_bytes(s, buff, l) == -1) {
	    perror("Could not write to socket");
	    close(l);
	    exit(1);
	}
    }
    if(nextsong) {
	flush_remote(s);
	nextsong = 0;
    }
    if(l == -1) { /* an error */
	fprintf(stderr, "Error reading %s", filename);
	perror("");
    }
    close(l);
}

int main(int argc, char *argv[]) {
    int s;
    struct sockaddr_in addr;
    char opt;
    int randomize = 0;
    size_t mtu = 8192;

    s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(s == -1) {
	perror("Failed to create socket");
	exit(1);
    }
    if(setsockopt(s, SOL_SOCKET, SO_SNDBUF, &mtu, sizeof(size_t)) == -1) {
	perror("Error initializing socket");
	exit(1);
    }
    addr.sin_family = AF_INET;
    addr.sin_port = htons(LISTENPORT);
    addr.sin_addr.s_addr = (u_int32_t) htonl(INADDR_LOOPBACK);

    if(connect(s, (struct sockaddr *) &addr, sizeof(struct sockaddr_in)) == -1) {
	perror("Failed to connect to localhost:5555");
	exit(1);
    }

    while((opt = (char)getopt(argc, argv, "zh")) != -1) {
	switch(opt) {
	case 'z':
	    randomize = 1;
	    break;
	case 'h':
	    usage(argv[0]);
	    exit(0);
	    break;
	default:
	    usage(argv[0]);
	    exit(1);
	}
    }

    if(randomize) {
	shuffle_files(argv + optind, argc - optind);
    }

    signal(SIGINT, sigintfunc);
    signal(SIGALRM, sigalarmfunc);

    while(optind <= argc) {
	stream_file(s, argv[optind]);
	optind++;
    }
}
